import os
import re
import numpy as np
import pandas as pd
from tqdm import tqdm


def from_mist(path):
    with open(path, 'r') as f:
        lines = f.readlines()
    
    init_names1 = lines[3].strip('#\n').split()
    init_values1 = lines[4].strip('#\n').split()
    init_names2 = lines[6].strip('#\n').split()
    init_values2 = lines[7].strip('#\n').split()

    initial_met = float(init_values1[init_names1.index('[Fe/H]')])
    initial_mass = float(init_values2[init_names2.index('initial_mass')])

    columns = lines[11].strip('#\n').split()
    data = np.genfromtxt(lines[12:])

    s = np.arange(len(data))
    m = np.ones_like(s) * initial_mass
    z = np.ones_like(s) * initial_met

    # Build multi-indexed DataFrame, dropping unwanted columns
    multi_index = pd.MultiIndex.from_tuples(zip(m, z, s),
        names=['initial_mass', 'initial_met', 'eep'])
    df = pd.DataFrame(data, index=multi_index, columns=columns)

    return df

def all_from_mist(raw_grids_path, progress=True):
    filelist = []
    for folder in os.listdir(raw_grids_path):
        path = os.path.join(raw_grids_path, folder)
        files = [os.path.join(path, f) for f in os.listdir(path) if '.eep' in f]
        filelist += files

    df_list = []

    if progress:
        file_iter = tqdm(filelist)
    else:
        file_iter = filelist

    df_list = []
    for fname in file_iter:
        try:
            df_list.append(from_mist(fname))
        except:
            print(f'Error reading {fname}---skipping.')
    dfs = pd.concat(df_list).sort_index()

    return dfs    

def install(
    raw_grids_path,
    name=None,
    ):
    '''
    The main method to install grids that are output of the `rotevol` rotational
    evolution tracer code.

    Parameters
    ----------
    raw_grids_path (str): the path to the folder containing the raw model grids.

    name (str, optional): the name of the grid you're installing. By default,
        the basename of the `raw_grids_path` will be used.

    Returns None
    '''
    from .stargrid import from_pandas
    from .stargrid import grids_path as install_path

    if name is None:
        name = os.path.basename(raw_grids_path)

    # Create cache directories
    path = os.path.join(install_path, name)
    if not os.path.exists(path):
        os.makedirs(path)

    eeps = all_from_mist(raw_grids_path)
    eeps = from_pandas(eeps, name=name)

    # Save EEP grid to file
    eep_save_path = os.path.join(path, 'eep_grid.pqt')
    print(f'Saving to {eep_save_path}')
    eeps.to_parquet(eep_save_path)

    # Create and save interpolator to file
    interp = eeps.to_interpolator()
    interp_save_path = os.path.join(path, 'interpolator.pkl')
    print(f'Saving interpolator to {interp_save_path}')
    interp.to_pickle(path=interp_save_path)

    print(f'Model grid "{name}" installed.')


'''
interp.py
from Tim Morton's isochrones package:
https://github.com/timothydmorton/isochrones

Copyright (c) 2014 timothydmorton

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
'''

import os
import itertools

import numba as nb
from math import sqrt
import numpy as np
import pandas as pd


@nb.jit(nopython=True)
def searchsorted(arr, x, N=-1):
    """N is length of arr
    """
    if N == -1:
        N = len(arr)

    L = 0
    R = N - 1
    done = False
    eq = False
    m = (L + R) // 2
    while not done:
        xm = arr[m]
        if xm < x:
            L = m + 1
        elif xm > x:
            R = m - 1
        elif xm == x:
            L = m
            eq = True
            done = True
        m = (L + R) // 2
        if L > R:
            done = True
    return L, eq


@nb.jit(nopython=True)
def find_indices(point, iis):
    ndim = len(point)

    indices = np.zeros(ndim, dtype=nb.uint32)
    norm_distances = np.zeros(ndim, dtype=nb.float64)
    out_of_bounds = False
    for i in range(ndim):
        ii = iis[i]
        n = len(ii)
        x = point[i]
        ix, eq = searchsorted(ii, x)
        if eq:
            indices[i] = ix
            norm_distances[i] = 0
        else:
            ix = ix - 1
            indices[i] = ix
            dx = ii[ix + 1] - ii[ix]
            norm_distances[i] = (x - ii[ix]) / dx
        out_of_bounds &= x < ii[0] or x > ii[n - 1]

    return indices, norm_distances, out_of_bounds


@nb.jit(nopython=True)
def find_indices_2d(x0, x1, ii0, ii1):

    n0 = len(ii0)
    n1 = len(ii1)

    indices = np.empty(2, dtype=nb.uint32)
    norm_distances = np.empty(2, dtype=nb.float64)

    if (x0 < ii0[0]) or (x0 > ii0[n0 - 1]) or (x1 < ii1[0]) or (x1 > ii1[n1 - 1]):
        return indices, norm_distances, True  # Out of bounds

    ix, eq = searchsorted(ii0, x0)
    if eq:
        indices[0] = ix
        norm_distances[0] = 0
    else:
        indices[0] = ix - 1
        c0 = ii0[ix - 1]
        norm_distances[0] = (x0 - c0) / (ii0[ix] - c0)

    ix, eq = searchsorted(ii1, x1)
    if eq:
        indices[1] = ix
        norm_distances[1] = 0
    else:
        indices[1] = ix - 1
        c0 = ii1[ix - 1]
        norm_distances[1] = (x1 - c0) / (ii1[ix] - c0)

    return indices, norm_distances, False


@nb.jit(nopython=True)
def find_indices_3d(x0, x1, x2, ii0, ii1, ii2):

    n0 = len(ii0)
    n1 = len(ii1)
    n2 = len(ii2)

    indices = np.empty(3, dtype=nb.uint32)
    norm_distances = np.empty(3, dtype=nb.float64)

    if (
        (x0 < ii0[0])
        or (x0 > ii0[n0 - 1])
        or (x1 < ii1[0])
        or (x1 > ii1[n1 - 1])
        or (x2 < ii2[0])
        or (x2 > ii2[n2 - 1])
    ):
        return indices, norm_distances, True  # Out of bounds

    ix, eq = searchsorted(ii0, x0)
    if eq:
        indices[0] = ix
        norm_distances[0] = 0
    else:
        indices[0] = ix - 1
        c0 = ii0[ix - 1]
        norm_distances[0] = (x0 - c0) / (ii0[ix] - c0)

    ix, eq = searchsorted(ii1, x1)
    if eq:
        indices[1] = ix
        norm_distances[1] = 0
    else:
        indices[1] = ix - 1
        c0 = ii1[ix - 1]
        norm_distances[1] = (x1 - c0) / (ii1[ix] - c0)

    ix, eq = searchsorted(ii2, x2)
    if eq:
        indices[2] = ix
        norm_distances[2] = 0
    else:
        indices[2] = ix - 1
        c0 = ii2[ix - 1]
        norm_distances[2] = (x2 - c0) / (ii2[ix] - c0)

    return indices, norm_distances, False


@nb.jit(nopython=True)
def find_indices_4d(x0, x1, x2, x3, ii0, ii1, ii2, ii3):

    n0 = len(ii0)
    n1 = len(ii1)
    n2 = len(ii2)
    n3 = len(ii3)

    indices = np.empty(4, dtype=nb.uint32)
    norm_distances = np.empty(4, dtype=nb.float64)

    if (
        (x0 < ii0[0])
        or (x0 > ii0[n0 - 1])
        or (x1 < ii1[0])
        or (x1 > ii1[n1 - 1])
        or (x2 < ii2[0])
        or (x2 > ii2[n2 - 1])
        or (x3 < ii3[0])
        or (x3 > ii3[n3 - 1])
    ):
        return indices, norm_distances, True  # Out of bounds

    ix, eq = searchsorted(ii0, x0)
    if eq:
        indices[0] = ix
        norm_distances[0] = 0
    else:
        indices[0] = ix - 1
        c0 = ii0[ix - 1]
        norm_distances[0] = (x0 - c0) / (ii0[ix] - c0)

    ix, eq = searchsorted(ii1, x1)
    if eq:
        indices[1] = ix
        norm_distances[1] = 0
    else:
        indices[1] = ix - 1
        c0 = ii1[ix - 1]
        norm_distances[1] = (x1 - c0) / (ii1[ix] - c0)

    ix, eq = searchsorted(ii2, x2)
    if eq:
        indices[2] = ix
        norm_distances[2] = 0
    else:
        indices[2] = ix - 1
        c0 = ii2[ix - 1]
        norm_distances[2] = (x2 - c0) / (ii2[ix] - c0)

    ix, eq = searchsorted(ii3, x3)
    if eq:
        indices[3] = ix
        norm_distances[3] = 0
    else:
        indices[3] = ix - 1
        c0 = ii3[ix - 1]
        norm_distances[3] = (x3 - c0) / (ii3[ix] - c0)

    return indices, norm_distances, False


@nb.jit(nopython=True)
def interp_value_2d(x0, x1, grid, icols, ii0, ii1):
    if x0 != x0 or x1 != x1:
        return np.array([np.nan for i in icols])

    indices, norm_distances, out_of_bounds = find_indices_2d(x0, x1, ii0, ii1)

    if out_of_bounds:
        return np.array([np.nan for i in icols])
    # The following should be equivalent to
    #  edges = np.array(list(itertools.product(*[[i, i+1] for i in indices])))

    ndim = 2
    n_edges = 2 ** ndim
    edges = np.zeros((n_edges, ndim))
    for i in range(n_edges):
        for j in range(ndim):
            edges[i, j] = indices[j] + ((i >> (ndim - 1 - j)) & 1)  # woohoo!

    n_values = len(icols)
    values = np.zeros(n_values, dtype=nb.float64)

    for j in range(n_edges):
        edge_indices = np.zeros(ndim, dtype=nb.uint32)
        for k in range(ndim):
            edge_indices[k] = edges[j, k]

        weight = 1.0
        for ei, i, yi in zip(edge_indices, indices, norm_distances):
            if ei == i:
                weight *= 1 - yi
            else:
                weight *= yi

        for i_icol in range(n_values):
            icol = icols[i_icol]

            # Now, get the value; this is why general ND doesn't work
            grid_indices = (edge_indices[0], edge_indices[1], icol)
            values[i_icol] += grid[grid_indices] * weight

    return values


@nb.jit(nopython=True)
def interp_value_3d(x0, x1, x2, grid, icols, ii0, ii1, ii2):
    if x0 != x0 or x1 != x1 or x2 != x2:
        return np.array([np.nan for i in icols])

    indices, norm_distances, out_of_bounds = find_indices_3d(x0, x1, x2, ii0, ii1, ii2)

    if out_of_bounds:
        return np.array([np.nan for i in icols])
    # The following should be equivalent to
    #  edges = np.array(list(itertools.product(*[[i, i+1] for i in indices])))

    ndim = 3
    n_edges = 2 ** ndim
    edges = np.zeros((n_edges, ndim))
    for i in range(n_edges):
        for j in range(ndim):
            edges[i, j] = indices[j] + ((i >> (ndim - 1 - j)) & 1)  # woohoo!

    n_values = len(icols)
    values = np.zeros(n_values, dtype=nb.float64)

    for j in range(n_edges):
        edge_indices = np.zeros(ndim, dtype=nb.uint32)
        for k in range(ndim):
            edge_indices[k] = edges[j, k]

        weight = 1.0
        for ei, i, yi in zip(edge_indices, indices, norm_distances):
            if ei == i:
                weight *= 1 - yi
            else:
                weight *= yi

        for i_icol in range(n_values):
            icol = icols[i_icol]

            # Now, get the value; this is why general ND doesn't work
            grid_indices = (edge_indices[0], edge_indices[1], edge_indices[2], icol)
            values[i_icol] += grid[grid_indices] * weight

    return values


@nb.jit(nopython=True)
def interp_value_4d(x0, x1, x2, x3, grid, icols, ii0, ii1, ii2, ii3):
    if x0 != x0 or x1 != x1 or x2 != x2 or x3 != x3:
        return np.array([np.nan for i in icols])

    indices, norm_distances, out_of_bounds = find_indices_4d(x0, x1, x2, x3, ii0, ii1, ii2, ii3)

    if out_of_bounds:
        return np.array([np.nan for i in icols])

    # The following should be equivalent to
    #  edges = np.array(list(itertools.product(*[[i, i+1] for i in indices])))

    ndim = 4
    n_edges = 2 ** ndim
    edges = np.zeros((n_edges, ndim))
    for i in range(n_edges):
        for j in range(ndim):
            edges[i, j] = indices[j] + ((i >> (ndim - 1 - j)) & 1)  # woohoo!

    n_values = len(icols)
    values = np.zeros(n_values, dtype=nb.float64)

    for j in range(n_edges):
        edge_indices = np.zeros(ndim, dtype=nb.uint32)
        for k in range(ndim):
            edge_indices[k] = edges[j, k]

        weight = 1.0
        for ei, i, yi in zip(edge_indices, indices, norm_distances):
            if ei == i:
                weight *= 1 - yi
            else:
                weight *= yi

        for i_icol in range(n_values):
            icol = icols[i_icol]

            # Now, get the value; this is why general ND doesn't work
            grid_indices = (edge_indices[0], edge_indices[1], edge_indices[2], edge_indices[3], icol)
            values[i_icol] += grid[grid_indices] * weight

    return values


@nb.jit(nopython=True)
def interp_values_2d(xx0, xx1, grid, icols, ii0, ii1):
    """xx1, xx2, xx3 are all arrays at which values are desired
    """

    N = len(xx0)
    ncols = len(icols)
    results = np.empty((N, ncols), dtype=nb.float64)
    for i in range(N):
        res = interp_value_2d(xx0[i], xx1[i], grid, icols, ii0, ii1)
        for j in range(ncols):
            results[i, j] = res[j]

    return results


@nb.jit(nopython=True)
def interp_values_3d(xx0, xx1, xx2, grid, icols, ii0, ii1, ii2):
    """xx1, xx2, xx3 are all arrays at which values are desired
    """

    N = len(xx0)
    ncols = len(icols)
    results = np.empty((N, ncols), dtype=nb.float64)
    for i in range(N):
        res = interp_value_3d(xx0[i], xx1[i], xx2[i], grid, icols, ii0, ii1, ii2)
        for j in range(ncols):
            results[i, j] = res[j]

    return results


@nb.jit(nopython=True)
def interp_values_4d(xx0, xx1, xx2, xx3, grid, icols, ii0, ii1, ii2, ii3):
    """xx1, xx2, xx3 are all arrays at which values are desired
    """

    N = len(xx0)
    ncols = len(icols)
    results = np.empty((N, ncols), dtype=nb.float64)
    for i in range(N):
        res = interp_value_4d(xx0[i], xx1[i], xx2[i], xx3[i], grid, icols, ii0, ii1, ii2, ii3)
        for j in range(ncols):
            results[i, j] = res[j]

    return results


@nb.jit(nopython=True)
def sign(x):
    if x < 0:
        return -1
    else:
        return 1


# @jit(nopython=True)
def find_closest3(
    val, a, b, v1, v2, grid, icol, ii1, ii2, ii3, bisect_tol=0.5, newton_tol=0.01, max_iter=100, debug=False
):
    """Find value of 3rd index array where interp_value is closest to val
    val : value to match
    a, b: min and max x-value to serch
=   x1, x2 : first and second values to pass to inter_values
    grid : 4d grid
    icol : index of value dimension of grid
    ii1, ii2, ii3 : grid dimension arrays
    """

    # First, do a bisect search to get it close
    done = False
    ya = interp_value_3d(v1, v2, a, grid, icol, ii1, ii2, ii3) - val
    yb = interp_value_3d(v1, v2, b, grid, icol, ii1, ii2, ii3) - val
    if debug:
        print("Initial values: {}: {}".format((a, b), (ya, yb)))

    if yb != yb or ya != ya:
        # bounds are nan, return nan.
        return np.nan
    elif abs(ya) < newton_tol:
        return float(a)
    elif abs(yb) < newton_tol:
        return float(b)
    elif ya > 0 and yb > 0:
        return np.nan
    elif yb < 0 and yb < 0:
        return np.nan

    else:
        if debug:
            print("doing bisect search...")

        while not done:
            c = (a + b) / 2
            yc = interp_value_3d(v1, v2, c, grid, icol, ii1, ii2, ii3) - val
            if yc == 0 or (b - a) / 2 < bisect_tol:
                done = True
            if sign(yc) == sign(ya):  # (yc >= 0 and ya >= 0) or (yc < 0 and ya < 0):
                a = c
                ya = yc
            else:
                b = c
                yb = yc
            if debug:
                print("{0} {1}".format((a, b, c), (ya, yb, yc)))

    # Now, use the value at index c to seed Newton-secant algorithm
    tol = 1000.0
    i = 0
    x0 = c
    y0 = yc
    x1 = x0 + 0.1
    y1 = interp_value_3d(v1, v2, x1, grid, icol, ii1, ii2, ii3) - val

    if debug:
        print("Newton-secant method...")
    while tol > newton_tol and i < max_iter:
        newx = (x0 * y1 - x1 * y0) / (y1 - y0)
        x0 = x1
        y0 = y1
        x1 = newx
        y1 = interp_value_3d(v1, v2, x1, grid, icol, ii1, ii2, ii3) - val

        # Boo!
        while not y1 == y1:
            if debug:
                print("{0} {1}".format(x1, y1))
            raise RuntimeError("ran into nan." + "run {} with debug=True to see why.".format((val, v2, v1)))

        if y1 >= 0:
            tol = y1
        else:
            tol = -y1
        i += 1
        if debug:
            print("{0} {1}".format(x1, y1))

    return x1


@nb.jit(nopython=True)
def interp_eeps(xs, x0s, x1s, ii0, ii1, n1, arrays, weight_arrays, lengths):
    n = len(xs)
    results = np.empty(n, dtype=nb.float64)

    for i in range(n):
        x = xs[i]
        x0 = x0s[i]
        x1 = x1s[i]
        results[i] = interp_eep(x, x0, x1, ii0, ii1, n1, arrays, weight_arrays, lengths)

    return results


class DFInterpolator(object):
    """Interpolate column values of DataFrame with full-grid hierarchical index
    """

    def __init__(self, df, filename=None, recalc=False, is_full=False):

        self.filename = filename
        self.is_full = is_full
        self.columns = list(df.columns)
        self.n_columns = len(self.columns)
        self.grid = self._make_grid(df, recalc=recalc)
        self.index_columns = tuple(np.array(l, dtype=float) for l in df.index.levels)
        self.index_names = df.index.names

        self.ndim = len(self.index_columns)

        self.column_index = {c: self.columns.index(c) for c in self.columns}

    def _make_grid(self, df, recalc=False):
        if self.filename is not None and os.path.exists(self.filename) and not recalc:
            d = np.load(self.filename)
            grid = d["grid"]
            columns = d["columns"]
            if not all(columns == self.columns):
                raise ValueError("DataFrame columns do not match columns loaded from full grid!")
        else:
            if not self.is_full:  # Need to make a full grid and pad with nans
                idx = pd.MultiIndex.from_tuples([ixs for ixs in itertools.product(*df.index.levels)])

                # Make an empty dataframe with the completely gridded index, and fill
                grid_df = pd.DataFrame(index=idx, columns=df.columns)
                grid_df.loc[df.index] = df
            else:
                grid_df = df

            shape = [len(l) for l in df.index.levels] + [len(df.columns)]

            grid = np.array(grid_df.values, dtype=float).reshape(shape)

            if self.filename is not None:
                np.savez(self.filename, grid=grid, columns=self.columns)

        return grid

    def add_column(self, values, name):
        newgrid = np.empty((self.grid.shape[:-1]) + (self.n_columns + 1,))
        newgrid[..., :-1] = self.grid
        newgrid[..., -1] = values
        self.column_index[name] = self.n_columns
        self.n_columns += 1
        self.columns += [name]
        self.grid = newgrid

    def find_closest(self, val, lo, hi, v1, v2, col="initial_mass", debug=False):
        icol = self.column_index[col]

        if self.ndim == 3:
            return find_closest3(val, lo, hi, v1, v2, self.grid, icol, *self.index_columns, debug=debug)

    def __call__(self, p, cols="all"):
        if cols == "all":
            icols = np.arange(self.n_columns)
        else:
            icols = np.array([self.column_index[col] for col in cols])
        args = (p, self.grid, icols, self.index_columns)

        if self.ndim == 2:
            args = (p[0], p[1], self.grid, icols, self.index_columns[0], self.index_columns[1])
            if (isinstance(p[0], float) or isinstance(p[0], int)) and (
                isinstance(p[1], float) or isinstance(p[1], int)
            ):
                values = interp_value_2d(*args)
            else:
                b = np.broadcast(*p)
                pp = [np.atleast_1d(np.resize(x, b.shape)).astype(float) for x in p]
                args = (*pp, self.grid, icols, *self.index_columns)
                # print([(a, type(a)) for a in args])
                values = interp_values_2d(*args)
        if self.ndim == 3:
            args = (
                p[0],
                p[1],
                p[2],
                self.grid,
                icols,
                self.index_columns[0],
                self.index_columns[1],
                self.index_columns[2],
            )
            if (
                (isinstance(p[0], float) or isinstance(p[0], int))
                and (isinstance(p[1], float) or isinstance(p[1], int))
                and (isinstance(p[2], float) or isinstance(p[2], int))
            ):
                values = interp_value_3d(*args)
            else:
                b = np.broadcast(*p)
                pp = [np.atleast_1d(np.resize(x, b.shape)).astype(float) for x in p]
                args = (*pp, self.grid, icols, *self.index_columns)
                # print([(a, type(a)) for a in args])
                values = interp_values_3d(*args)
        elif self.ndim == 4:
            args = (
                p[0],
                p[1],
                p[2],
                p[3],
                self.grid,
                icols,
                self.index_columns[0],
                self.index_columns[1],
                self.index_columns[2],
                self.index_columns[3],
            )
            if (
                (isinstance(p[0], float) or isinstance(p[0], int))
                and (isinstance(p[1], float) or isinstance(p[1], int))
                and (isinstance(p[2], float) or isinstance(p[2], int))
                and (isinstance(p[3], float) or isinstance(p[3], int))
            ):
                values = interp_value_4d(*args)
            else:
                b = np.broadcast(*p)
                pp = [np.atleast_1d(np.resize(x, b.shape)).astype(float) for x in p]
                values = interp_values_4d(*pp, self.grid, icols, *self.index_columns)

        return values


'''
progress_bar.py
from David Maxson's miniutils package:
https://github.com/scnerd/miniutils

MIT License

Copyright (c) 2017 David Maxson

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

---

Copied over when a compatibility issue with pyparsing and pycontracts broke 
miniutils. kiauhoku depends only on the two progress bars defined in this file, 
so this removes non-standard dependencies from kiauhoku.
'''

import itertools
import multiprocessing as mp
try:
    from nose.plugins.multiprocess import TimedOutException
except ImportError:
    TimedOutException = TimeoutError
import random
import warnings

try:
    from tqdm import tqdm as _tqdm
    try:  # pragma: nocover
        # Check if we're in a Jupyter notebook... if so, use the ipywidgets progress bar instead
        from IPython import get_ipython
        if type(get_ipython()).__module__.startswith('ipykernel.'):
            from tqdm import tqdm_notebook as _tqdm
    except (ImportError, NameError):  # pragma: nocover
        # IPython isn't even installed, or we're not in it
        pass
except ImportError:  # pragma: nocover
    # noinspection PyUnusedLocal
    def _tqdm(iterable, *a, **kw):
        return iterable


def progbar(iterable, *a, verbose=True, **kw):
    """Prints a progress bar as the iterable is iterated over

    :param iterable: The iterator to iterate over
    :param a: Arguments to get passed to tqdm (or tqdm_notebook, if in a Jupyter notebook)
    :param verbose: Whether or not to print the progress bar at all
    :param kw: Keyword arguments to get passed to tqdm
    :return: The iterable that will report a progress bar
    """
    iterable = range(iterable) if isinstance(iterable, int) else iterable
    if verbose:
        return _tqdm(iterable, *a, **kw)
    else:
        return iterable


def _fun(f, q_in, q_out, flatten, star):  # pragma: no cover
    try:
        while True:
            i, x = q_in.get()
            if i is None:
                break
            out = f(*x) if star else f(x)
            if flatten:
                for j, o in enumerate(out):
                    q_out.put(((i, j), o))
                q_out.put((None, None))
            else:
                q_out.put((i, out))
    except BaseException as ex:
        q_out.put((None, ex))


def _parallel_progbar_launch(mapper, iterable, nprocs=None, starmap=False, flatmap=False, shuffle=False,
                             verbose=True, verbose_flatmap=None, max_cache=-1, timeout=1, **kwargs):

    # Shuffle the iterable if requested, to make the parallel execution potentially more uniform in runtime
    enumerated_iterable = enumerate(iterable)
    if shuffle:
        enumerated_iterable = list(enumerated_iterable)
        # ids = [i for i in sorted(range(len(iterable)), key=lambda x: random.random())]
        # iterable = (iterable[i] for i in ids)
        random.shuffle(enumerated_iterable)  # Is this going to be expensive for large lists of large objects?

    # Check that we don't launch more processes than there are elements to map (if that's knowable)
    nprocs = nprocs or mp.cpu_count()
    try:
        nprocs = max(1, min(len(iterable), nprocs))
    except TypeError:
        pass

    # Set up multiprocessing management for mapping
    q_in = mp.Queue()
    q_out = mp.Queue(max_cache)

    procs = [mp.Process(target=_fun, args=(mapper, q_in, q_out, flatmap, starmap)) for _ in range(nprocs)]
    for p in procs:
        p.daemon = True
        p.start()

    # Doing it this way prevents us from storing locally an entire list of the input values unnecessarily, and still
    # gets us the number of elements sent for processing
    sent = (q_in.put((i, x)) for i, x in enumerated_iterable)
    num_sent = sum(1 for _ in sent)
    for _ in range(nprocs):
        # Send out a flag for each process to terminate once all elements are processed
        q_in.put((None, None))

    # Fetch the mapped results from the output queue, printing a progress bar as you go
    if flatmap:
        # If we're flat mapping, then we'll keep separate progress of all returned results (an unknown number) and how
        # many inputs are complete (a known number). The outer loop will track the latter, and the inner loop the former
        results = (q_out.get() for _ in progbar(itertools.count(),
                                                verbose=verbose if verbose_flatmap is None else verbose_flatmap,
                                                **kwargs))
        for _ in progbar(num_sent, verbose=verbose):
            for i, x in results:
                # When we're flagged that an input is done being returned in the queue, break the inner loop to make
                # the "completed inputs" progress bar tick
                if i is None:
                    if x is None:
                        break
                    else:
                        raise x
                yield i, x
    else:
        for i, x in (q_out.get() for _ in progbar(num_sent, verbose=verbose, **kwargs)):
            if i is None:
                raise x
            yield i, x

    # Clean up
    for p in procs:
        try:
            p.join(timeout)
        except (TimeoutError, mp.TimeoutError, TimedOutException):  # pragma: nocover
            warnings.warn("parallel_progbar mapping process failed to close properly (check error output)")


def parallel_progbar(*args, **kwargs):
    """Performs a parallel mapping of the given iterable, reporting a progress bar as values get returned

    :param mapper: The mapping function to apply to elements of the iterable
    :param iterable: The iterable to map
    :param nprocs: The number of processes (defaults to the number of cpu's)
    :param starmap: If true, the iterable is expected to contain tuples and the mapper function gets each element of a
        tuple as an argument
    :param flatmap: If true, flatten out the returned values if the mapper function returns a list of objects
    :param shuffle: If true, randomly sort the elements before processing them. This might help provide more uniform
        runtimes if processing different objects takes different amounts of time.
    :param verbose: Whether or not to print the progress bar
    :param verbose_flatmap: If performing a flatmap, whether or not to report each object as it's returned
    :param timeout: The number of seconds to wait for each worker process after completing
    :param kwargs: Any other keyword arguments to pass to the progress bar (see ``progbar``)
    :return: A list of the returned objects, in the same order as provided
    """

    results = _parallel_progbar_launch(*args, **kwargs)
    return [x for i, x in sorted(results, key=lambda p: p[0])]



import os
import pickle
import numpy as np
import pandas as pd
from tqdm import tqdm


# Assign labels used in eep conversion
eep_params = dict(
    age = 'Age(Gyr)',
    core_hydrogen_frac = 'Xcen',
    lum = 'Log L/Lsun',
    logg = 'logg',
    core_helium_frac = 'Ycen',
    teff_scale = 20, # used in metric function
    lum_scale = 1, # used in metric function
    # `intervals` is a list containing the number of secondary Equivalent
    # Evolutionary Phases (EEPs) between each pair of primary EEPs.
    intervals = [200, # Between PreMS and ZAMS
                  50, # Between ZAMS and EAMS 
                 100, # Between EAMS and IAMS
                 100, # IAMS-TAMS
                 150], # TAMS-RGBump
)

def my_PreMS(track, eep_params, i0=None):
    '''
    Garstec models do not have central temperature, which is necessary for
    the default PreMS calculation. For now, let the first point be the PreMS.
    '''
    return 0

def my_RGBump(track, eep_params, i0=None):
    '''
    Modified from eep.get_RGBump to make lum logarithmic
    Note that even though Teff is linear (as opposed to log in default usage),
    The actual value of Teff isn't used here--only comparative values--so I
    don't bother logging/unlogging Teff.
    '''

    lum = eep_params['lum']

    N = len(track)

    lum_tr = track.loc[i0:, lum]
    logT_tr = track.loc[i0:, 'Teff']

    lum_greater = (lum_tr > 1)
    if not lum_greater.any():
        return -1
    RGBump = lum_greater.idxmax() + 1

    while logT_tr[RGBump] < logT_tr[RGBump-1] and RGBump < N-1:
        RGBump += 1

    # Two cases: 1) We didn't reach an extremum, in which case RGBump gets
    # set as the final index of the track. In this case, return -1.
    # 2) We found the extremum, in which case RGBump gets set
    # as the index corresponding to the extremum.
    if RGBump >= N-1:
        return -1
    return RGBump-1

def my_HRD(track, eep_params):
    '''
    Adapted from eep._HRD_distance to fix lum and teff logarithms
    '''

    # Allow for scaling to make changes in Teff and L comparable
    Tscale = eep_params['teff_scale']
    Lscale = eep_params['lum_scale']

    lum = eep_params['lum']

    logTeff = np.log10(track['Teff'])
    logLum = track[lum]

    N = len(track)
    dist = np.zeros(N)
    for i in range(1, N):
        temp_dist = (((logTeff.iloc[i] - logTeff.iloc[i-1])*Tscale)**2
                    + ((logLum.iloc[i] - logLum.iloc[i-1])*Lscale)**2)
        dist[i] = dist[i-1] + np.sqrt(temp_dist)

    return dist

def read_columns(path):
    with open(path, 'r') as f:
        columns = [l.strip() for l in f]
    
    return columns

def parse_filename(filename):
    file_str = filename.replace('.col_mst', '')

    mass = int(file_str[3:6])/100

    met_str = file_str[7:11]
    met = int(met_str[1:])/100
    if met_str[0] == 'm':
        met *= -1

    return mass, met


def from_garstec(path, columns=None):
    if columns is None:
        raw_grids_path = os.path.dirname(path)
        columns = read_columns(os.path.join(raw_grids_path, 'column_labels.txt'))

    fname = os.path.basename(path)
    initial_mass, initial_met = parse_filename(fname)

    data = np.loadtxt(path)
    s = np.arange(len(data))
    m = np.ones_like(s) * initial_mass
    z = np.ones_like(s) * initial_met

    # Build multi-indexed DataFrame, dropping unwanted columns
    multi_index = pd.MultiIndex.from_tuples(zip(m, z, s),
        names=['initial_mass', 'initial_met', 'step'])
    df = pd.DataFrame(data, index=multi_index, columns=columns)
    df = df.drop(columns=[c for c in columns if '#' in c])

    return df

def all_from_garstec(raw_grids_path, progress=True):
    df_list = []
    filelist = [f for f in os.listdir(raw_grids_path) if '.col_mst' in f]
    columns = read_columns(os.path.join(raw_grids_path, 'column_labels.txt'))

    if progress:
        file_iter = tqdm(filelist)
    else:
        file_iter = filelist

    for fname in file_iter:
        fpath = os.path.join(raw_grids_path, fname)
        df_list.append(from_garstec(fpath, columns))

    dfs = pd.concat(df_list).sort_index()

    return dfs 

def install(
    raw_grids_path,
    name=None,
    eep_params=eep_params,
    eep_functions={'prems': my_PreMS, 'rgbump': my_RGBump},
    metric_function=my_HRD,
    ):
    '''
    The main method to install grids that are output of the `rotevol` rotational
    evolution tracer code.

    Parameters
    ----------
    raw_grids_path (str): the path to the folder containing the raw model grids.

    name (str, optional): the name of the grid you're installing. By default,
        the basename of the `raw_grids_path` will be used.

    eep_params (dict, optional): contains a mapping from your grid's specific
        column names to the names used by kiauhoku's default EEP functions.
        It also contains 'eep_intervals', the number of secondary EEPs
        between each consecutive pair of primary EEPs. By default, the params
        defined at the top of this script will be used, but users may specify
        their own.

    eep_functions (dict, optional): if the default EEP functions won't do the
        job, you can specify your own and supply them in a dictionary.
        EEP functions must have the call signature
        function(track, eep_params), where `track` is a single track.
        If none are supplied, the default functions will be used.

    metric_function (callable, None): the metric function is how the EEP
        interpolator spaces the secondary EEPs. By default, the path
        length along the evolution track on the H-R diagram (luminosity vs.
        Teff) is used, but you can specify your own if desired.
        metric_function must have the call signature
        function(track, eep_params), where `track` is a single track.
        If no function is supplied, defaults to garstec.my_HRD.

    Returns None
    '''
    from .stargrid import from_pandas
    from .stargrid import grids_path as install_path

    if name is None:
        name = os.path.basename(raw_grids_path)

    # Create cache directories
    path = os.path.join(install_path, name)
    if not os.path.exists(path):
        os.makedirs(path)

    # Cache eep parameters
    with open(os.path.join(path, 'eep_params.pkl'), 'wb') as f:
        pickle.dump(eep_params, f)

    print('Reading and combining grid files')
    grids = all_from_garstec(raw_grids_path)
    grids = from_pandas(grids, name=name)

    # Save full grid to file
    full_save_path = os.path.join(path, 'full_grid.pqt')
    print(f'Saving to {full_save_path}')
    grids.to_parquet(full_save_path)

    print(f'Converting to eep-based tracks')
    eeps = grids.to_eep(eep_params, eep_functions, metric_function)

    # Save EEP grid to file
    eep_save_path = os.path.join(path, 'eep_grid.pqt')
    print(f'Saving to {eep_save_path}')
    eeps.to_parquet(eep_save_path)

    # Create and save interpolator to file
    interp = eeps.to_interpolator()
    interp_save_path = os.path.join(path, 'interpolator.pkl')
    print(f'Saving interpolator to {interp_save_path}')
    interp.to_pickle(path=interp_save_path)

    print(f'Model grid "{name}" installed.')


import os
import re
import pickle
import numpy as np
import pandas as pd
from tqdm import tqdm


# Assign labels used in eep conversion
eep_params = dict(
    age = 'Age(Gyr)',
    log_central_temp = 'logT(cen)',
    core_hydrogen_frac = 'Xcen',
    hydrogen_lum = 'H lum (Lsun)',
    lum = 'L/Lsun',
    logg = 'logg',
    log_teff = 'Log Teff(K)',
    core_helium_frac = 'Ycen',
    teff_scale = 20, # used in metric function
    lum_scale = 1, # used in metric function
    # `intervals` is a list containing the number of secondary Equivalent
    # Evolutionary Phases (EEPs) between each pair of primary EEPs.
    intervals = [200, # Between PreMS and ZAMS
                  50, # Between ZAMS and EAMS 
                 100, # Between EAMS and IAMS
                 100, # IAMS-TAMS
                 150], # TAMS-RGBump
)


def from_rotevol(path):
    fname = path.split('/')[-1]
    val_str = fname.replace('.out', '')
    _, met_str, _, alpha_str = val_str.split('_')
    met = float(met_str)/100
    alpha = float(alpha_str)/100
        
    with open(path, 'r') as f:
        header = f.readline()
        # Header format: ` NUMBER OF TRACKS NNN ...`
        # Each file contains `ntracks` evolutionary tracks, each with `nsteps` steps
        
        ntracks = int(header[18:21])
        nsteps = np.zeros(ntracks, dtype=int)
        
        # Initial mass and period also specified in preamble
        mass_init = np.zeros(ntracks, dtype=float)
        period_init = np.zeros(ntracks, dtype=float)
        
        # Read preamble. First column is an unnecessart index.
        for i in range(ntracks):
            line = f.readline().split()
            _, nsteps[i], mass_init[i], period_init[i] = line
            
        # Read the column label line
        columns = re.split(r'\s{2,}', f.readline().strip())
        
        # Begin reading tracks
        df_tracks = []
        for i in range(ntracks):
            # Construct track array
            track = np.zeros((nsteps[i], len(columns)))
            for j in range(nsteps[i]):
                track[j] = f.readline().split()

            # Convert track to DataFrame and save.
            df_track = pd.DataFrame(track, columns=columns)
            df_tracks.append(df_track)

    df_tracks = pd.concat(df_tracks)
    j_track = df_tracks['J'].values - 1
    k_step = df_tracks['K'].values.astype('int') - 1
    m = np.round(0.3 + j_track/100, 2)
    z = np.round(np.ones_like(m) * met, 1)
    a = np.round(np.ones_like(m) * alpha, 1)
    s = k_step

    # Build multi-indexed DataFrame, dropping unwanted columns
    multi_index = pd.MultiIndex.from_tuples(zip(m, z, a, s),
        names=['initial_mass', 'initial_met', 'initial_alpha', 'step'])
    df = pd.DataFrame(df_tracks.values, index=multi_index, columns=columns)
    df = df.drop(columns=['J', 'K', 'dummyGamma', 'dummyCcore'])
    df = df.loc[:, ~df.columns.duplicated()]

    return df

def all_from_rotevol(raw_grids_path, progress=True):
    df_list = []
    filelist = [f for f in os.listdir(raw_grids_path) if '.out' in f]

    if progress:
        file_iter = tqdm(filelist)
    else:
        file_iter = filelist

    for fname in file_iter:
        fpath = os.path.join(raw_grids_path, fname)
        df_list.append(from_rotevol(fpath))

    dfs = pd.concat(df_list).sort_index()

    return dfs    

def install(
    raw_grids_path,
    name=None,
    eep_params=eep_params,
    eep_functions=None,
    metric_function=None,
    ):
    '''
    The main method to install grids that are output of the `rotevol` rotational
    evolution tracer code.

    Parameters
    ----------
    raw_grids_path (str): the path to the folder containing the raw model grids.

    name (str, optional): the name of the grid you're installing. By default,
        the basename of the `raw_grids_path` will be used.

    eep_params (dict, optional): contains a mapping from your grid's specific
        column names to the names used by kiauhoku's default EEP functions.
        It also contains 'eep_intervals', the number of secondary EEPs
        between each consecutive pair of primary EEPs. By default, the params
        defined at the top of this script will be used, but users may specify
        their own.

    eep_functions (dict, optional): if the default EEP functions won't do the
        job, you can specify your own and supply them in a dictionary.
        EEP functions must have the call signature
        function(track, eep_params), where `track` is a single track.
        If none are supplied, the default functions will be used.

    metric_function (callable, None): the metric function is how the EEP
        interpolator spaces the secondary EEPs. By default, the path
        length along the evolution track on the H-R diagram (luminosity vs.
        Teff) is used, but you can specify your own if desired.
        metric_function must have the call signature
        function(track, eep_params), where `track` is a single track.
        If no function is supplied, defaults to kiauhoku.eep._HRD_distance.

    Returns None
    '''
    from .stargrid import from_pandas
    from .stargrid import grids_path as install_path

    if name is None:
        name = os.path.basename(raw_grids_path)

    # Create cache directories
    path = os.path.join(install_path, name)
    if not os.path.exists(path):
        os.makedirs(path)

    # Cache eep parameters
    with open(os.path.join(path, 'eep_params.pkl'), 'wb') as f:
        pickle.dump(eep_params, f)

    print('Reading and combining grid files')
    grids = all_from_rotevol(raw_grids_path)
    grids = from_pandas(grids, name=name)

    # Save full grid to file
    full_save_path = os.path.join(path, 'full_grid.pqt')
    print(f'Saving to {full_save_path}')
    grids.to_parquet(full_save_path)

    print(f'Converting to eep-based tracks')
    eeps = grids.to_eep(eep_params, eep_functions, metric_function)

    # Save EEP grid to file
    eep_save_path = os.path.join(path, 'eep_grid.pqt')
    print(f'Saving to {eep_save_path}')
    eeps.to_parquet(eep_save_path)

    # Create and save interpolator to file
    interp = eeps.to_interpolator()
    interp_save_path = os.path.join(path, 'interpolator.pkl')
    print(f'Saving interpolator to {interp_save_path}')
    interp.to_pickle(path=interp_save_path)

    print(f'Model grid "{name}" installed.')


import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
import pandas as pd

import kiauhoku as kh


def my_hist(data, title, eep_indices):
    fig = plt.figure()
    ax = fig.add_subplot(111)
    
    fs = 14
    fontdict = {"fontsize": fs}
    labeldict = {"labelsize": fs}

    ax.imshow(data, aspect="auto", cmap=my_cmap(), norm=my_norm(eep_indices), 
              extent=(0.295, 2.005, 0.75, -1.25), picker=True)
    ax.invert_yaxis()
    ax.set_title(title, **fontdict)
    ax.set_ylabel("[M/H]", **fontdict)
    ax.set_xlabel(r"$M$ / $M_\odot$", **fontdict)
    ax.set_xticks(np.linspace(0.305, 1.995, 170), minor=True)
    ax.set_xticks(np.linspace(0.3, 2.0, 18), minor=False)
    ax.set_yticks(np.linspace(-1.25, 0.75, 5), minor=True)
    ax.set_yticks(np.linspace(-1.0, 0.5, 4), minor=False)
    ax.xaxis.grid(True, which="minor", color="gray", linestyle="--")
    ax.yaxis.grid(True, which="minor", color="k")
    ax.tick_params(axis="both", which="minor", length=0)
    ax.tick_params(axis="both", which="major", **labeldict)

    textdict = {"rotation": 90, 
                "fontsize": 18,
                "horizontalalignment": "center",
                "verticalalignment": "center"}

    ax.text(0.8, -1.0, "RGBump", **textdict)
    ax.text(0.66,-1.0, "TAMS", **textdict)
    ax.arrow(0.63, -1.0, -0.02, 0, zorder=5, 
             width=0.002, head_width=0.05, head_length=0.02, fc="k")
    ax.text(0.53, -1.0, "IAMS", **textdict)
    ax.text(0.40, -1.0, "EAMS", **textdict)
    ax.text(0.32, -1.0, "ZAMS", **textdict)

    return fig, ax


def my_cmap():
    clist = ["black", "salmon", "sandybrown", "khaki", "mediumseagreen", "skyblue"]
    cmap = mpl.colors.ListedColormap(clist)
    return cmap

def my_norm(eep_indices):
    bounds = [-1] + [i+2 for i in eep_indices]
    norm = mpl.colors.BoundaryNorm(bounds, my_cmap().N)
    return norm


def length_distribution(grid0, grid1, verbose=True):
    lens = []
    n_len = []
    for i in grid0:
        for j in i:
            if j not in lens:
                lens.append(j)
                n_len.append(1)
            else:
                idx = lens.index(j)
                n_len[idx] += 1
    for i in grid1:
        for j in i:
            if j not in lens:
                lens.append(j)
                n_len.append(1)
            else:
                idx = lens.index(j)
                n_len[idx] += 1

    if verbose:
        print("Lengths:                ", lens)
        print("Number of each length:  ", n_len)
        print("Total number of tracks: ", sum(n_len))

    return sorted(lens)


def measure_lengths(grid_name):
    eep_grid = kh.load_eep_grid(grid_name)
    eep_lengths = eep_grid.get_eep_track_lengths()

    metallicities = [-1.0, -0.5, 0.0, 0.5]
    masses = np.linspace(30, 200, 171)/100
    grid0 = np.zeros((4, 171), dtype=int)
    grid1 = np.zeros((4, 171), dtype=int)

    for i, z in enumerate(metallicities):
        for j, m in enumerate(masses):
            try:
                grid0[i, j] = eep_lengths.loc[(m, z, 0), 0]
            except:
                pass
            try:
                grid1[i, j] = eep_lengths.loc[(m, z, 0.4), 0]
            except:
                pass
    
    lens = length_distribution(grid0, grid1)
    return grid0, grid1


def make_plots(grid0, grid1, eep_indices):
    f1, a1 = my_hist(grid0, r"[$\alpha$/M] = 0.0", eep_indices)
    f2, a2 = my_hist(grid1, r"[$\alpha$/M] = 0.4", eep_indices)
    #f1.canvas.mpl_connect("button_press_event", onclick000)
    #f2.canvas.mpl_connect("button_press_event", onclick040)
    xlim = a1.get_xlim()
    ylim = a1.get_ylim()

    #a1.plot(data["mass"], data["met"], "ko", ms=3)
    #a2.plot(data["mass"], data["met"], "ko", ms=3)
    a1.set_xlim(xlim)
    a1.set_ylim(ylim)
    a2.set_xlim(xlim)
    a2.set_ylim(ylim)

    plt.show()

'''
def onclick000(event):
    try:
        mass = round(event.xdata, 2)
        met = round(2*(event.ydata))/2
        print("Mass: %.2f Msun | [M/H]: %.1f" %(mass, met))
        alpha = 0
        _HRD_model(mass, met, alpha)
    except TypeError:
        pass


def onclick040(event):
    try:
        mass = round(event.xdata, 2)
        met = round(2*(event.ydata))/2
        print("Mass: %.2f Msun | [M/H]: %.1f" %(mass, met))
        alpha = 0.4
        _HRD_model(mass, met, alpha)
    except TypeError:
        pass
'''

if __name__ == "__main__":
    name = 'fastlaunch'
    eep_params = kh.stargrid.load_eep_params(name)
    intervals = eep_params['intervals']
    eep_indices = np.arange(len(intervals) + 1)
    eep_indices[1:] += np.cumsum(intervals)

    grid0, grid1 = measure_lengths(name)
    make_plots(grid0, grid1, eep_indices)



name = "kiauhoku"

from .version import __version__
from .stargrid import install_grid, load_interpolator
from .stargrid import load_grid, load_full_grid, load_eep_grid

from . import rotevol, mist, yrec, garstec, dartmouth
from . import eep


'''
custom_install.py

This is a template to help users install custom grids to kiauhoku. The template
is based on the installation for YREC grids.

Necessary:
(1) a function called `setup` that returns a pandas MultiIndexed 
    DataFrame containing all your evolution tracks
(2) a variable `name` that is set to whatever 
    you want your installed grid to be named
(3) a variable `raw_grids_path` that sets the path to where
    your raw grid is downloaded.

Optional:
 - You can customize the functions you want to use to define your EEPs
   (see my_RGBump for an example). You'll need to group them in a dictionary
   called `eep_functions` that gets imported by kiauhoku. The allowed keys are
    'prems' for Pre-Main Sequence
    'zams'  for Zero-Age Main Sequence
    'eams'  for Early-Age Main Sequence
    'iams'  for Intermediate-Age Main Sequence
    'tams'  for Terminal-Age Main Sequence
    'rgbump' for the Red Giant Branch Bump (~ halfway up to the TRGB)
 - You can customize the metric function you use to space the EEPs
   (see my_HRD for an example). It either must be called `metric_function`
   or be aliased as `metric_function`.
'''

import os
import pickle
import numpy as np
import pandas as pd
from tqdm import tqdm

name = 'custom_yrec_example'
raw_grids_path = '/home/zach/Downloads/grids/yrec'

# Assign labels used in eep conversion
eep_params = dict(
    age = 'Age(Gyr)',
    log_central_temp = 'logT(cen)',
    core_hydrogen_frac = 'Xcen',
    hydrogen_lum = 'H lum (Lsun)',
    lum = 'L/Lsun',
    logg = 'logg',
    log_teff = 'Log Teff(K)',
    core_helium_frac = 'Ycen',
    teff_scale = 20, # used in metric function
    lum_scale = 1, # used in metric function
    # `intervals` is a list containing the number of secondary Equivalent
    # Evolutionary Phases (EEPs) between each pair of primary EEPs.
    intervals = [200, # Between PreMS and ZAMS
                  50, # Between ZAMS and EAMS 
                 100, # Between EAMS and IAMS
                 100, # IAMS-TAMS
                 150], # TAMS-RGBump
)

def my_RGBump(track, eep_params, i0=None):
    '''
    Modified from eep.get_RGBump to make luminosity logarithmic
    '''

    lum = eep_params['lum']
    log_teff = eep_params['log_teff']
    N = len(track)

    lum_tr = track.loc[i0:, lum]
    logT_tr = track.loc[i0:, log_teff]

    lum_greater = (lum_tr > 1)
    if not lum_greater.any():
        return -1
    RGBump = lum_greater.idxmax() + 1

    while logT_tr[RGBump] < logT_tr[RGBump-1] and RGBump < N-1:
        RGBump += 1

    # Two cases: 1) We didn't reach an extremum, in which case RGBump gets
    # set as the final index of the track. In this case, return -1.
    # 2) We found the extremum, in which case RGBump gets set
    # as the index corresponding to the extremum.
    if RGBump >= N-1:
        return -1
    return RGBump-1

def my_HRD(track, eep_params):
    '''
    Adapted from eep._HRD_distance to fix lum logarithm
    '''

    # Allow for scaling to make changes in Teff and L comparable
    Tscale = eep_params['teff_scale']
    Lscale = eep_params['lum_scale']

    log_teff = eep_params['log_teff']
    lum = eep_params['lum']

    logTeff = track[log_teff]
    logLum = track[lum]

    N = len(track)
    dist = np.zeros(N)
    for i in range(1, N):
        temp_dist = (((logTeff.iloc[i] - logTeff.iloc[i-1])*Tscale)**2
                    + ((logLum.iloc[i] - logLum.iloc[i-1])*Lscale)**2)
        dist[i] = dist[i-1] + np.sqrt(temp_dist)

    return dist

eep_functions = {
    'rgbump': my_RGBump
}

metric_function = my_HRD

def read_columns(path):
    with open(path, 'r') as f:
        columns = [l.strip() for l in f]
    
    return columns

def parse_filename(filename):
    file_str = filename.replace('.track', '')

    mass = float(file_str[:4].replace('_', '.'))

    met_i = file_str.find('fh') + 2
    met_str = file_str[met_i:met_i+4]
    met = float(met_str[1:])/100
    if  met != 0 and met_str[0] == 'm':
        met *= -1

    alpha_i = file_str.find('al') + 2
    alpha_str = file_str[alpha_i:alpha_i+2]
    alpha = float(alpha_str)/10

    return mass, met, alpha


def from_yrec(path, columns=None):
    if columns is None:
        raw_grids_path = os.path.dirname(path)
        columns = read_columns(os.path.join(raw_grids_path, 'column_labels.txt'))

    fname = os.path.basename(path)
    initial_mass, initial_met, initial_alpha = parse_filename(fname)

    data = np.loadtxt(path)
    s = np.arange(len(data))
    m = np.ones_like(s) * initial_mass
    z = np.ones_like(s) * initial_met

    # Build multi-indexed DataFrame, dropping unwanted columns
    multi_index = pd.MultiIndex.from_tuples(zip(m, z, s),
        names=['initial_mass', 'initial_met', 'step'])
    df = pd.DataFrame(data, index=multi_index, columns=columns)
    df = df.drop(columns=[c for c in columns if '#' in c])

    return df

def setup(raw_grids_path=raw_grids_path, progress=True):
    df_list = []
    filelist = [f for f in os.listdir(raw_grids_path) if '.track' in f]
    columns = read_columns(os.path.join(raw_grids_path, 'column_labels.txt'))

    if progress:
        file_iter = tqdm(filelist)
    else:
        file_iter = filelist

    for fname in file_iter:
        fpath = os.path.join(raw_grids_path, fname)
        df_list.append(from_yrec(fpath, columns))

    dfs = pd.concat(df_list).sort_index()
    # If you want to compute a total hydrogen luminosity, uncomment the next line
    #dfs[eep_params['hydrogen lum']] = dfs[['ppI', 'ppII', 'ppIII']].sum(axis=1)

    return dfs 


import setuptools

# Load the __version__ variable without importing the package already
exec(open("kiauhoku/version.py").read())

setuptools.setup(
    name="kiauhoku",
    version=__version__,
    author="Zachary R. Claytor",
    author_email="zclaytor@hawaii.edu",
    description="Utilities for interacting and interpolating stellar evolution models",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/zclaytor/kiauhoku",
    license="MIT",
    python_requires='>=3',
    install_requires=[
        'tqdm', 'numpy', 'pandas', 'scipy', 'emcee', 'pyarrow', 'numba'
    ],
    packages=setuptools.find_packages(),
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Astronomy",
    ],
)



'''
eep.py

Utilities to downsample stellar evolution tracks to Equivalent Evolutionary
Phase (EEP) basis, according to the method of Dotter (2016). 

The default EEP functions are contained in eep.default_eep_functions. They are
    default_eep_functions = {
        'prems': get_PreMS,
        'zams': get_ZAMS,
        'eams': get_EAMS,
        'iams': get_IAMS,
        'tams': get_TAMS,
        'rgbump': get_RGBump,
    }

You can define and supply your own EEP functions in a dictionary. 
EEP functions must have the call signature
    function(track, eep_params)
where `track` is a single track.
'''

import numpy as np
import pandas as pd
from scipy.interpolate import interp1d


def _eep_interpolate(track, eep_params, eep_functions, metric_function=None):
    '''
    Given a raw evolutionary track, returns a downsampled track based on
    Equivalent Evolutionary Phases (EEPs). The primary EEPs are defined below,
    and the default ones to use in the computation are listed in
    eep.default_eep_functions. The secondary EEPs are computed based on the
    number of secondary EEPs between each pair of primary EEPs as specified
    in `eep_params.intervals`; these should be defined in the grid
    installation file. If one of the EEP_intervals is 200, then for that pair
    of primary EEPs, the metric distance between those primary EEPs is divided
    into 200 equally spaced points, and the relevant stellar parameters are
    linearly interpolated at those points.

    Parameters
    ----------
    track (StarGrid): single-index StarGrid to be interpolated.

    eep_params (dict): dictionary of column names to use in EEP computation, 
        as well as secondary EEP intervals.

    eep_functions (dict): dictionary of callables to use for EEP computation.

    metric_function (callable, optional): function to compute EEP intervals.
        If none is specified, eep._HRD_distance will be used.

    Returns
    -------
    eep_track, a pandas DataFrame containing the EEP-based track.
    '''

    i_eep = _locate_primary_eeps(track, eep_params, eep_functions)
    num_intervals = len(i_eep) - 1
    # In some cases, the raw models do not hit the ZAMS. In these cases,
    # return None.
    if num_intervals == 0:
        return

    if metric_function is None:
        metric_function = _HRD_distance

    # compute metric distance along track
    dist = metric_function(track, eep_params)

    primary_eep_dist = dist[i_eep]
    eep_intervals = eep_params['intervals']
    secondary_eep_dist = np.zeros(
        sum(eep_intervals[:num_intervals]) + len(i_eep)
    )

    # Determine appropriate distance to each secondary EEP
    j0 = 0
    for i in range(num_intervals):
        my_dist = primary_eep_dist[i+1] - primary_eep_dist[i]
        delta = my_dist/(eep_intervals[i] + 1)
        new_dist = np.array([primary_eep_dist[i] + delta*j \
                for j in range(eep_intervals[i]+1)])
        
        secondary_eep_dist[j0:j0+len(new_dist)] = new_dist
        j0 += len(new_dist)

    secondary_eep_dist[-1] = primary_eep_dist[-1]

    # Create list of interpolator functions
    interp = interp1d(dist, track.T)

    # Interpolate stellar parameters along evolutionary track for
    # desired EEP distances
    eep_track = pd.DataFrame(interp(secondary_eep_dist).T, columns=track.columns)
    eep_track.index.name = 'eep'

    return eep_track

def _locate_primary_eeps(track, eep_params, eep_functions=None):
    '''
    Given a track, returns a list containing indices of Equivalent
    Evolutionary Phases (EEPs)
    '''

    # define a list of functions to iterate over
    eep_f = default_eep_functions
    if eep_functions is None:
        eep_functions = default_eep_functions
    else:
        eep_f.update(eep_functions)        

    # get indices of EEPs
    i_eep = []
    i_start = 0
    for f in eep_f.values():
        i_phase = f(track, eep_params, i0=i_start)
        i_eep.append(i_phase)
        i_start = i_phase
        if i_start == -1:
            return np.array(i_eep[:-1])
    
    return np.array(i_eep)
    
def get_PreMS(track, eep_params, i0=0, logTc_crit=5.0):
    '''
    The pre-main sequence EEP is the point where central temperature rises
    above a certain value (which must be lower than necessary for sustained
    fusion). The default value is log10(T_c) = 5.0, but may be chosen to be
    different. An optional argument i0 can be supplied, which is the
    index to start with.

    This relies on the behavior of pandas.Series.idxmax() for a Series
    of bools. If no temperature is greater than or equal to logTc, the 
    natural return value is i0. So we don't mistake this failed search,
    we must check the value at i0 to make sure it satisfies our criterion.

    Returns
    -------
    `i_PreMS`: (int) index of the first element in track[i0: "logT(cen)"]
    greater than logTc.    
    '''
    log_central_temp = eep_params['log_central_temp']

    logTc_tr = track.loc[i0:, log_central_temp]
    i_PreMS = _first_true_index(logTc_tr >= logTc_crit)
    return i_PreMS

def get_ZAMS(track, eep_params, i0=10, ZAMS_pref=3, Xc_burned=0.001, 
    Hlum_frac_max=0.999):
    '''
    The Zero-Age Main Sequence EEP has three different implementations in 
    Dotter's code:
    ZAMS1) the point where the core hydrogen mass fraction has been depleted
            by some fraction (0.001 by default: Xc <= Xmax - 0.001)
    ZAMS2) the point *before ZAMS1* where the hydrogen-burning luminosity 
            achieves some fraction of the total luminosity 
            (0.999 by default: Hlum/lum = 0.999)
    ZAMS3) the point *before ZAMS1* with the highest surface gravity

    ZAMS3 is implemented by default.
    '''
    core_hydrogen_frac = eep_params['core_hydrogen_frac']
    Xc_init = track.at[0, core_hydrogen_frac]
    Xc_tr = track.loc[i0:, core_hydrogen_frac]
    ZAMS1 = _first_true_index(Xc_tr <= Xc_init-Xc_burned)

    if ZAMS1 == -1:
        return -1

    if ZAMS_pref == 1:
        return ZAMS1

    if ZAMS_pref == 2:
        hydrogen_lum = eep_params['hydrogen_lum']
        lum = eep_params['lum']

        Hlum_tr = track.loc[i0:ZAMS1, hydrogen_lum]
        lum_tr = track.loc[i0:ZAMS1, lum]
        Hlum_frac = Hlum_tr/lum_tr
        ZAMS2 = _first_true_index(Hlum_frac >= Hlum_frac_max)
        if ZAMS2 == -1:
            return ZAMS1
        return ZAMS2

    # or ZAMS_pref = 3
    logg = eep_params['logg']
    logg_tr = track.loc[0:ZAMS1, logg]
    ZAMS3 = logg_tr.idxmax()
    return ZAMS3

def get_IorT_AMS(track, eep_params, i0, Xmin):
    '''
    The Intermediate- and Terminal-Age Main Sequence (IAMS, TAMS) EEPs both use
    the core hydrogen mass fraction dropping below some critical amount.
    This function encapsulates the main part of the code, with the difference
    between IAMS and TAMS being the value of Xmin.
    '''
    core_hydrogen_frac = eep_params['core_hydrogen_frac']
    Xc_tr = track.loc[i0:, core_hydrogen_frac]
    i_eep = _first_true_index(Xc_tr <= Xmin)
    return i_eep 

def get_EAMS(track, eep_params, i0=12, Xmin=0.55):
    '''
    Early-Age Main Sequence. Without this, the low-mass rotevol tracks do not
    reach an EEP past the ZAMS before 15 Gyr.
    '''
    i_EAMS = get_IorT_AMS(track, eep_params, i0, Xmin)
    return i_EAMS

def get_IAMS(track, eep_params, i0=12, Xmin=0.3):
    '''
    Intermediate-Age Main Sequence exists solely to ensure the convective
    hook is sufficiently sampled.
    Defined to be when the core hydrogen mass fraction drops below some
    critical value. Default: Xc <= 0.3
    '''
    i_IAMS = get_IorT_AMS(track, eep_params, i0, Xmin)
    return i_IAMS

def get_TAMS(track, eep_params, i0=14, Xmin=1e-12):
    '''
    Terminal-Age Main Sequence, defined to be when the core hydrogen mass
    fraction drops below some critical value. Default: Xc <= 1e-12
    '''
    i_TAMS = get_IorT_AMS(track, eep_params, i0, Xmin)
    return i_TAMS

def get_RGBump(track, eep_params, i0=None):
    '''
    The Red Giant Bump is an interruption in the increase in luminosity on the 
    Red Giant Branch. It occurs when the hydrogen-burning shell reaches the
    composition discontinuity left from the first convective dredge-up.

    Dotter skips the Red Giant Bump and proceeds to the Tip of the Red Giant
    Branch, but since the YREC models, at the time of this writing, terminate
    at the helium flash, I choose to use the Red Giant Bump as my final EEP.

    I identify the RGBump as the first local minimum in Teff after the TAMS.
    To avoid weird end-of-track artifacts, if the minimum is within 1 step
    from the end of the raw track, the track is treated as if it doesn't reach
    the RGBump.

    Added 2018/07/22: Some tracks have weird, jumpy behavior before the RGBump
    which gets mistakenly identified as the RGBump. To avoid this, I force the
    RGBump to be the first local minimum in Teff after the TAMS *and* with
    a luminosity above 10 Lsun.
    '''

    lum = eep_params['lum']
    log_teff = eep_params['log_teff']
    N = len(track)

    lum_tr = track.loc[i0:, lum]
    logT_tr = track.loc[i0:, log_teff]

    RGBump = _first_true_index(lum_tr > 10) + 1
    if RGBump == 0:
        return -1

    while logT_tr[RGBump] < logT_tr[RGBump-1] and RGBump < N-1:
        RGBump += 1

    # Two cases: 1) We didn't reach an extremum, in which case RGBump gets
    # set as the final index of the track. In this case, return -1.
    # 2) We found the extremum, in which case RGBump gets set
    # as the index corresponding to the extremum.
    if RGBump >= N-1:
        return -1
    return RGBump-1

def get_RGBTip(track, eep_params, i0=None):
    '''
    Red Giant Branch Tip
    Dotter describes the tip of the red giant branch (RGBTip) EEP as
    "the point at which stellar luminosity reaches a maximum---or the stellar
    Teff reaches a minimum---after core H burning is complete but before core
    He burning has progressed significantly."

    Note that the YREC models at the time of this writing nominally end at
    the helium flash, so the RGBTip is not recommended to use as an EEP.
    '''

    core_helium_frac = eep_params['core_helium_frac']
    lum = eep_params['lum']
    log_teff = eep_params['log_teff']

    Ymin = track.at[i0, core_helium_frac] - 1e-2
    Yc_tr = track.loc[i0:, core_helium_frac]
    before_He_burned = (Yc_tr > Ymin)
    if not before_He_burned.any():
        return -1

    lum_tr = track.loc[i0:, lum]
    RGBTip1 = (lum_tr[before_He_burned]).idxmax()

    logT_tr = track.loc[i0:, log_teff]
    RGBTip2 = (logT_tr[before_He_burned]).idxmin()

    RGBTip = min(RGBTip1, RGBTip2)
    return RGBTip

def get_ZACHeB(track, eep_params, i0=None):
    '''
    Zero-Age Core Helium Burning
    Denotes the onset of sustained core He burning. The point is identified 
    as the core temperature minimum that occurs after the onset of He burning
    (RGBTip) while Ycen > Ycen,RGBTip - 0.03.
    '''

    core_helium_frac = eep_params['core_helium_frac']
    log_central_temp = eep_params['log_central_temp']
        
    Ymin = track.at[i0, core_helium_frac] - 0.03
    Yc_tr = track.loc[i0:, core_helium_frac]
    before_He_burned = (Yc_tr > Ymin)
    
    if not before_He_burned.any():
        return -1

    logTc_tr = track.loc[i0:, log_central_temp]
    ZACHeB = (logTc_tr[before_He_burned]).idxmin()

    return ZACHeB

def get_TACHeB(track, eep_params, i0=None, Yc_crit=1e-4):
    '''
    Terminal-Age Core Helium Burning
    Corresponds to Ycen = 1e-4.
    '''

    core_helium_frac = eep_params['core_helium_frac']
    Yc_tr = track.loc[i0:, core_helium_frac]

    TACHeB = _first_true_index(Yc_tr < Yc_crit)
    return TACHeB

def get_TPAGB(track, eep_params, i0=None):
    '''
    Thermally-Pulsing Asymptotic Giant Branch
    Not yet implemented.
    '''

    raise NotImplementedError(
        'Function "get_TPAGB" not yet implemented.'
    )

def get_PostAGB(track, eep_params, i0=None):
    '''
    Post AGB
    Not yet implemented.
    '''

    raise NotImplementedError(
        'Function "get_PostTPAGB" not yet implemented.'
    )

def get_WDCS(track, eep_params, i0=None):
    '''
    White Dwarf Cooling Sequence
    Not yet implemented.
    '''

    raise NotImplementedError(
        'Function "get_WDCS" not yet implemented.'
    )


def _HRD_distance(track, eep_params):
    '''
    The Metric Function is used to calculate the distance along the evolution
    track. Traditionally, the Euclidean distance along the track on the
    H-R diagram has been used, but any positive-definite function will work.
    
    Distance along the H-R diagram, to be used in the Metric Function.
    Returns an array containing the distance from the beginning of the 
    evolution track for each step along the track, in logarithmic effective
    temperature and logarithmic luminosity space.
    '''

    # Allow for scaling to make changes in Teff and L comparable
    Tscale = eep_params['teff_scale']
    Lscale = eep_params['lum_scale']

    log_teff = eep_params['log_teff']
    lum = eep_params['lum']

    logTeff = track[log_teff]
    logLum = np.log10(track[lum])

    N = len(track)
    dist = np.zeros(N)
    for i in range(1, N):
        temp_dist = (((logTeff.iloc[i] - logTeff.iloc[i-1])*Tscale)**2
                    + ((logLum.iloc[i] - logLum.iloc[i-1])*Lscale)**2)
        dist[i] = dist[i-1] + np.sqrt(temp_dist)

    return dist

def _first_true_index(bools):
    '''
    Given a pandas Series of bools, returns the index of the first occurrence
    of `True`. **Index-based, NOT location-based**
    I.e., say x = pd.Series({0: False, 2: False, 4: True}), then
    _first_true_index(x) will return index 4, not the positional index 2.

    If no value in `bools` is True, returns -1.
    '''
    if not bools.any():
        return -1
    return bools.idxmax()

default_eep_functions = {
    'prems': get_PreMS,
    'zams': get_ZAMS,
    'eams': get_EAMS,
    'iams': get_IAMS,
    'tams': get_TAMS,
    'rgbump': get_RGBump,
}


'''
stargrid.py

Contains classes and functions to interact with and interpolate stellar
evolutionary model grids
'''

import os
from importlib import import_module
import pickle
import functools

import numpy as np
import pandas as pd
from scipy.interpolate import interp1d
from scipy.optimize import minimize
import emcee

from .eep import _eep_interpolate
from .interp import DFInterpolator
from .progress_bar import progbar, parallel_progbar


grids_path = os.path.expanduser('~/') + '.kiauhoku/grids/'
interp_path = os.path.expanduser('~/') + '.kiauhoku/interpolators/'

class StarGrid(pd.DataFrame):
    '''
    StarGrid is designed to store and interact with stellar evolution tracks.
    It is little more than a pandas DataFrame with a few extra features.

    Parameters
    ----------
    name (str): the name of the grid, e.g., 'mist'

    eep_params (dict): for EEP-based grids, eep_params contains a mapping from
        your grid's specific column names to the names used by kiauhoku's
        default EEP functions. It also contains 'eep_intervals', the number
        of secondary EEPs between each consecutive pair of primary EEPs.
    '''

    def __init__(self, *args, **kwargs):
        name = kwargs.pop('name', None)
        eep_params = kwargs.pop('eep_params', None)

        # use the __init__ method from DataFrame to ensure
        # that we're inheriting the correct behavior
        super(StarGrid, self).__init__(*args, **kwargs)

        self._metadata = ['name', 'eep_params']
        # Set StarGrid name
        self.name = name or 'Anonymous StarGrid'
        self.eep_params = eep_params

    # this method makes it so our methods return an instance
    # of StarGrid instead of a regular DataFrame
    @property
    def _constructor(self):
        return StarGrid

    def __setattr__(self, attr, val):
        # have to special case custom attributes because
        # pandas tries to set as columns
        if attr in self._metadata:
            object.__setattr__(self, attr, val)
        else:
            super(StarGrid, self).__setattr__(attr, val)

    def set_name(self, name):
        '''Set the name of the grid.'''
        self.name = name

    @property
    def index_range(self):
        '''Returns the range of index columns.
        '''
        idx = self.index.droplevel(-1)
        mins = [idx.get_level_values(n).min() for n in idx.names]
        maxs = [idx.get_level_values(n).max() for n in idx.names]
        return pd.Series(zip(mins, maxs), index=idx.names)

    def get_track(self, index):
        '''Get a track at a specific index.

        Parameters
        ----------
        index (tuple): corresponds to the input indices of the grid. For
            example, if your grid is indexed by mass, metallicity, and alpha-
            abundance, to get the track for mass=1, metallicity=0.5, alpha=0,

            >>> track = grid.get_track((1, 0.5, 0))

        Returns
        -------
        track (StarGrid): desired evolution track, now indexed only by step.
        '''
        return self.loc[index, :]

    def is_MultiIndex(self):
        '''Checks whether the StarGrid instance is MultiIndexed.'''
        return isinstance(self.index, pd.MultiIndex)

    def to_eep(self,
        eep_params=None,
        eep_functions=None,
        metric_function=None,
        progress=True,
        nprocs=None,
        **kwargs
    ):
        '''
        Converts the grid of evolution tracks to EEP basis. For details on EEP
        functions, see the documentation for kiauhoku.eep.

        Parameters
        ----------
        eep_params (dict, None): contains a mapping from your grid's specific
            column names to the names used by kiauhoku's default EEP functions.
            It also contains 'eep_intervals', the number of secondary EEPs
            between each consecutive pair of primary EEPs. If none are supplied,
            kiauhoku will attempt to read them from a cache directory.

        eep_functions (dict, None): if the default EEP functions won't do the
            job, you can specify your own and supply them in a dictionary.
            EEP functions must have the call signature
            function(track, eep_params), where `track` is a single track.
            If none are supplied, the default functions will be used.

        metric_function (callable, None): the metric function is how the EEP
            interpolator spaces the secondary EEPs. By default, the path
            length along the evolution track on the H-R diagram (luminosity vs.
            Teff) is used, but you can specify your own if desired.
            metric_function must have the call signature
            function(track, eep_params), where `track` is a single track.
            If no function is supplied, defaults to kiauhoku.eep._HRD_distance.

        progress (bool, True): whether or not to display a progress bar.

        nprocs (int, None): how many parallel processes to use for MultiIndex
            DataFrames. If none is specified, defaults to the number of CPUs.

        **kwargs: extra keyword arguments to pass to parallel_progbar.

        Returns
        -------
        eep_frame (StarGrid): grid of EEP-based evolution tracks.
        '''

        # User can specify eep_params, but if none are specified,
        # searched for cached params.
        if not eep_params:
            eep_params = load_eep_params(self.name)

        # If self is a MultiIndexed DataFrame, split it into individual
        # tracks, convert to EEP basis, and recombine.
        if self.is_MultiIndex():
            partial_eep = functools.partial(_eep_pool_helper, self,
                eep_params, eep_functions, metric_function)

            # create index iterator and pass to the mapping/progress function
            idx = self.index.droplevel(-1).drop_duplicates()
            #tracks = self.groupby(level=list(range(idx.nlevels)))
            eep_tracks = parallel_progbar(partial_eep, idx, 
                verbose=progress, nprocs=nprocs, **kwargs)

            # Setup MultiIndex and remove Nones
            idx_list = [(*i, j) for i, tr in zip(idx, eep_tracks) if tr is not None for j in tr.index]
            eep_tracks = [tr for tr in eep_tracks if tr is not None]

            # Create MultiIndex for EEP frame
            multiindex = pd.MultiIndex.from_tuples(
                idx_list,
                names=[*idx.names, 'eep']
            )

            # Put it all together
            eep_frame = pd.concat(eep_tracks, ignore_index=True)
            eep_frame.index = multiindex

        # Other case is if a single track is passed
        else:
            eep_frame = _eep_interpolate(
                self, eep_params, eep_functions, metric_function
            )

        # Cast DataFrame to StarGrid
        eep_frame = from_pandas(eep_frame, name=self.name, eep_params=eep_params)

        return eep_frame

    def to_interpolator(self):
        '''Converts EEP-based grid to StarGridInterpolator.
        '''
        if 'eep' not in self.index.names:
            raise RuntimeError('Grid is wrong kind. Must be EEP grid.')
        return StarGridInterpolator(self)        

    def get_primary_eeps(self):
        '''Return indices of Primary EEPs in the EEP-based tracks.
        '''
        if 'eep' not in self.index.names:
            raise RuntimeError('Grid is wrong kind. Must be EEP grid.')

        ints = [0] + self.eep_params['intervals']
        eeps = np.arange(len(ints)) + np.cumsum(ints)
        
        return eeps

    def get_eep_track_lengths(self):
        '''
        This is mainly a convenience function to be used in the script
        `eep_track_lengths.py`, but that script is currently configured to work
        only for the rotevol grids.

        Returns
        -------
        lengths: pandas DataFrame containing the number of EEPs in each track
            of the grid.
        '''

        if 'eep' not in self.index.names:
            raise RuntimeError('Grid is wrong kind. Must be EEP grid.')

        idx = self.index.droplevel('eep').drop_duplicates()
        lengths = [len(self.loc[i]) for i in idx]
        lengths = pd.DataFrame(lengths, index=idx)
        return lengths

class StarGridInterpolator(DFInterpolator):
    '''
    Stellar model grid interpolator. Built on the DataFrame Interpolator
    (DFInterpolator) of Tim Morton's isochrones package, the
    StarGridInterpolator is intended to provide easy interaction with stellar
    model grids.

    Attributes
    ----------
    name (str): name of the grid

    columns (list-like): the available columns in the grid.

    max_eep (int): the maximum EEP index out of all the tracks.

    eep_params (dict): the parameters used to calculate the EEPs.
    '''

    def __init__(self, grid):
        super(StarGridInterpolator, self).__init__(grid)

        self.name = grid.name or 'Anonymous StarGridInterpolator'
        self.columns = grid.columns
        self.index = grid.index

        self.max_eep = grid.index.to_frame().eep.max()
        self.eep_params = grid.eep_params

    @property
    def index_range(self):
        '''Returns the range of index columns.
        '''
        idx = self.index.droplevel(-1)
        mins = [idx.get_level_values(n).min() for n in idx.names]
        maxs = [idx.get_level_values(n).max() for n in idx.names]
        return pd.Series(zip(mins, maxs), index=idx.names)

    def get_primary_eeps(self):
        '''Return indices of Primary EEPs in the EEP-based tracks.
        '''
        ints = [0] + self.eep_params['intervals']
        eeps = np.arange(len(ints)) + np.cumsum(ints)
        
        return eeps

    def get_star_eep(self, index):
        '''
        Interpolate a single model or list of models from the grid.
        Note that this is the preferred way to sample models from the grid.

        `index` should be a tuple of indices in the same way you would access
        a model from a StarGrid. If your grid is indexed by mass and
        metallicity, and you want the 350th EEP of a 0.987-solar-mass,
        0.2-metallicity star:

        >>> star = grid.get_star_eep((0.987, 0.2, 350))
        '''

        star_values = self(index)
        if len(np.shape(index)) == 1:
            star = pd.Series(star_values, index=self.columns)
        else:
            star = pd.DataFrame(star_values, columns=self.columns)

        return star

    def get_star_age(self, index, age, age_label=None):
        '''
        Interpolate a single model from the grid, accessing by age.
        Note that this method is slower than get_star_eep. get_star_age
        interpolates an entire track from the grid, then runs a 1-D
        interpolator over the track to get the parameters for the desired
        age. get_star_eep is preferred to this method.

        `index` should be a tuple of indices in the same way you would access
        a model from a StarGrid. If your grid is indexed by mass and
        metallicity, and you want a a 4.5-Gyr-old, 0.987-solar-mass,
        0.2-metallicity star:

        >>> star = grid.get_star_age((0.987, 0.2), 4.5)

        Optional Arguments
        ------------------
        age_label (str, None): ideally, you should specify what your grid calls
            age in eep_params in your setup file. If you did, then get_star_age
            can figure out what to call 'age' from the eep_params that are
            stored in the interpolator. If you didn't do this, you can specify
            what your grid calls age using age_label. If grid uses 'Age(Gyr)':

            >>> star = grid.get_star_age(
                    (0.987, 0.2), 4.5, age_label='Age(Gyr)'
                )
        '''

        track = self.get_track(index)
        labels = track.columns
        if age_label is None:
            eep_params = self.eep_params
            if eep_params is None:
                raise ValueError(
                    'No eep_params are stored. Please specify age_label.'
                )
            else:
                age_label = eep_params['age']

        interpf = interp1d(track[age_label], track.values.T)
        star = pd.Series(interpf(age), labels)
        return star

    def get_track(self, index):
        '''
        Interpolate a single track from the model grid.

        `index` should be a tuple of indices in the same way you would access
        a model from a StarGrid. If your grid is indexed by mass and
        metallicity, and you want a track for a 0.987-solar-mass,
        0.2-metallicity star:

        >>> star = grid.get_track((0.987, 0.2))
        '''

        num_eeps = self.max_eep + 1
        ones_arr = np.ones(num_eeps)
        idx = [i*ones_arr for i in index] + [np.arange(num_eeps)]
        star_values = self(idx)
        track = StarGrid(star_values, columns=self.columns,
                         name=self.name, eep_params=self.eep_params)
        return track

    def mcmc_star(self, log_prob_fn, args,
        pos0=None, initial_guess=None, guess_width=None,
        n_walkers=None, n_burnin=0, n_iter=500,
        save_path=None, **kwargs,
    ):
        '''
        Uses emcee to sample stellar models from the grid.
        For example usage, see mcmc.ipynb in the parent kiauhoku directory.

        Parameters
        ----------
        log_prob_fn (callable): the log-probability function to be passed
            to the emcee EnsembleSampler. Should have call signature
            log_prob_fn(pos, interp, ...), where `pos` is the walker position,
            `interp` is the StarGridInterpolator, and other arguments can be
            supplied as needed using `args`.
            log_prob_fn should return (1), the computed log-probability as a
            float, and (2), the sampled star model from the interpolator. This
            allows `blobs` to be used, and for you to keep track other stellar
            parameters not directly used in the sampling.
            See the docs for emcee for more advanced usage.

        args (tuple): extra arguments to be passed to log_prob_fn.

        pos0 (numpy ndarray, shape n_dim x n_walkers):
            You can optionally directly supply the EnsembleSampler the initial
            walker positions. Alternatively, you can supply a single walker
            position `initial_guess` and take `n_walkers` samples from a
            gaussian distribution with width `guess_width`.

        initial_guess (tuple, optional): initial walker position, to be sampled
            n_walkers times. Should be the same shape as the model grid index.
            Use as an alternative to `pos0`.

        guess_width (tuple, optional): width of initial guess sampling. Should
            be the same shape as the model grid index. Use as an alternative 
            to `pos0`.

        n_walkers (int, optional): number of walkers. If pos0 is specified,
            n_walkers is inferred from the shape of pos0. Otherwise,
            defaults to 12.

        n_burnin (int, optional): number of burn-in steps. Default: 0.

        n_iter (int, optional): number of sample steps. Default: 500.

        save_path (str, optional): You may optionally specify a path to a
            CSV or Parquet file to save the sampler output as a DataFrame.
            Use of Parquet requires that you have pyarrow or another parquet-
            compatible package installed.

        kwargs: Extra keyword arguments to pass to the EnsembleSampler.

        Returns
        -------
        sampler, the emcee.EnsembleSampler object

        output, a pandas DataFrame comprised of the flattened Markov chains
            from the sampler, plus all the stellar parameters returned from
            each interpolated sample.
        '''

        # If pos0 is not specified, construct it from initial_guess and width
        if pos0 is None:
            if n_walkers is None:
                n_walkers = 12

            pos0 = np.array([
                np.random.normal(guess, width, n_walkers)
                for guess, width in zip(initial_guess, guess_width)
            ]).T

        elif n_walkers is None:
            n_walkers = len(pos0)

        sampler = emcee.EnsembleSampler(
            n_walkers,
            len(initial_guess),
            log_prob_fn,
            args=(self, *args),
            blobs_dtype=[('star', pd.Series)],
            **kwargs,
        )

        # Run burn-in stage
        if n_burnin > 0:
            pos, prob, state, blobs = sampler.run_mcmc(pos0, n_burnin, progress=True)
            sampler.reset()
        else:
            pos = pos0

        # Run sampling stage
        pos, prob, state, blobs = sampler.run_mcmc(pos, n_iter, progress=True)

        samples = pd.DataFrame(sampler.flatchain, columns=self.index.names)
        blobs = sampler.get_blobs(flat=True)
        blobs = pd.concat(blobs['star'], axis=1).T

        # Concatenate Markov chains with blobs
        output = pd.concat([samples, blobs], axis=1)

        # Save output if desired
        if save_path:
            if 'csv' in save_path:
                output.to_csv(save_path, index=False)
            elif 'pqt' in save_path:
                output.to_parquet(save_path, index=False)
            else:
                print(
                    'save_path extension not recognized, so chains were not saved:\n'
                    f'    {save_path}\n'
                    'Accepted extensions are .csv and .pqt.'
                )

        return sampler, output

    def fit_star(self, star_dict, guess, *args,
                 loss='meansquarederror', scale=None, **kwargs
    ):
        '''
        Fit a star from data using `scipy.optimize.minimize`.

        Parameters
        ----------
        star_dict: dict containing label-value pairs for the star to be fit

        guess: tuple containing initial guess of input values for star.
            These should be of the same form as the input to
            `StarGridInterpolator.get_star_eep`.

        *args: extra arguments to be passed to the loss function.

        loss: string specifying the loss function to be used.
            'meansquarederror' and 'meanpercenterror' are implemented.
            Defaults to 'measquarederror'.

        scale: optional tuple of scale factors to be used in the
            meansquarederror computation. Defaults to None.
            If `scale` is specified with meanpercenterror loss, an
            error will be raised.

        **kwargs: extra keyword arguments to be passed to `scipy.optimize.minimize`.

        Returns
        -------
        result: the output of `scipy.optimize.minimize`.
        '''

        if loss in ['meansquarederror', 'mse']:
            loss_function = self._meansquarederror
        elif loss in ['meanpercenterror', 'meanpcterr', 'mpe']:
            loss_function = self._meanpercenterror
        else:
            raise NotImplementedError(
                f'Loss function {loss} not implemented.'
            )

        if scale is not None:
            args = (star_dict, scale, *args)
        else:
            args = (star_dict, *args)

        result = minimize(loss_function, guess, args=args, method='Nelder-Mead', **kwargs)

        return result

    def gridsearch_fit(self, star_dict, *args, scale=None, tol=1e-6,
                    mass_step=0.1, met_step=0.2, alpha_step=0.2, eep_step=50, 
                    verbose=True, **kwargs):
        '''
        Aggressively fit a star using `scipy.optimize.minimize` across the
        whole grid of models until a sufficient match is found.

        There are three possible cases:
        (1) A fit is found whose loss value is within `tol` tolerance. If this
            happens, the search ceases and the fit is returned.
        (2) `scipy.optimize.minimize` successfully identifies a fit, but it is
            not within the user-specified tolerance. In this case, the entire
            grid will be searched, and the best fit will be returned.
        (3) `scipy.optimize.minimize` fails converge to a solution. In this
            case, a `None` is returned with the most recent scipy output.

        Parameters
        ----------
        star_dict (dict): dictionary containing label-value pairs to be fit.

        *args: extra arguments to be passed to `StarGridInterpolator.fit_star`.

        scale (tuple, None): scale factors by which to divide the values of 
            star_dict to put them to the same order of magnitude. This speeds
            up the fitting process in test cases and also improves accuracy.

        tol (float, 1e-6): user-specified tolerance for the fit. The tolerance
            represents the desired value of the loss. If a solution is found
            within the tolerance, the gridsearch will cease. 

        mass_step (float, 0.1): the mass spacing between scipy optimizers.

        met_step (float, 0.2): the metallicity spacing between scipy optimizers.

        alpha_step (float, 0.2): the alpha-abundance spacing between scipy
            optimizers. If `initial_alpha` is not in the index, this is ignored.

        eep_step (float, 50): the EEP spacing between scipy optimizers.

        verbose (bool, True): whether to print fit messages. Recommended to
            leave as `True` unless you're running a large list of stars AND
            you know what you're doing.

        **kwargs: extra keyword arguments to be passed to `fit_star`.

        Returns
        -------
        best_model (pandas Series): the stellar parameters for the best fit, if
            a fit was achieved. Otherwise this will be `None`.

        best_fit (`scipy.optimize.optimize.OptimizeResult`): the scipy 
            optimizer result containing information pertaining to the fit.
        '''

        if verbose:
            print(f'Fitting star with {self.name}...')

        # Construct a multi-index iterator instead of using a nested for-loop
        idxrange = self.index_range
        idx_list = []
        if 'initial_mass' in idxrange:
            # * operator unravels the tuple to pass values to altrange
            mass_list = altrange(*idxrange["initial_mass"], mass_step)
            idx_list.append(mass_list)
        if 'initial_met' in idxrange:
            met_list = altrange(*idxrange['initial_met'], met_step)
            idx_list.append(met_list)
        if 'initial_alpha' in idxrange:
            alpha_list = altrange(*idxrange['initial_alpha'], alpha_step)
            idx_list.append(alpha_list)
        eep_list = np.arange(252, 606, eep_step)
        idx_list.append(eep_list)

        idx_list = pd.MultiIndex.from_product(idx_list)

        # Loop through indices searching for fit
        best_loss = 1e10
        some_fit = False
        good_fit = False
        for idx in idx_list:
            fit = self.fit_star(star_dict, idx, *args, scale=scale, **kwargs)
            if fit.success:
                some_fit = True
                if fit.fun < best_loss:
                    best_fit = fit
                    best_loss = fit.fun
                    if fit.fun <= tol:
                        good_fit = True
                        if verbose:
                            print(f'{self.name}: success!')
                        break

        # Check to see how the fit did, print comments if desired.
        if not some_fit:
            if verbose:
                print(f'*!*!*!* {self.name} fit failed! Returning last attempt.')
            return None, fit
        if verbose and not good_fit:
            print(f'{self.name}: Fit not converged to within tolerance, but returning closest fit.')

        # get the model, add the indices, and return
        fit_idx = best_fit.x
        best_model = self.get_star_eep(fit_idx)
        for label, value in zip(idxrange.index, fit_idx[:-1]):
            best_model[label] = value
        best_model['eep'] = fit_idx[-1]
        
        return best_model, best_fit

    def _meansquarederror(self, index, star_dict, scale=False):
        '''Mean Squared Error loss function for `fit_star`.

        Parameters
        ----------
        index (tuple): index to be fit.

        star_dict (dict): dictionary of values for loss function computation.

        scale (list-like, optional): Optionally scale the squared errors before
            taking the mean. This could be useful if, for example, luminosity is
            in solar units (~1) and age is in years (~10^9 years).

        Returns
        -------
        mean squared error as a float.
        '''

        star = self.get_star_eep(index)
        sq_err = np.array([(star[l] - star_dict[l])**2 for l in star_dict])

        if scale:
            sq_err /= np.array(scale)**2

        return np.average(sq_err)

    def _meanpercenterror(self, index, star_dict):
        '''Mean Percent Error loss function for `fit_star`.

        Parameters
        ----------
        index (tuple): index to be fit.

        star_dict (dict): dictionary of values for loss function computation.

        Returns
        -------
        mean percent error as a float.
        '''

        star = self.get_star_eep(index)
        mpe = np.average(
            [np.abs(star[l] - star_dict[l])/star_dict[l] for l in star_dict]
        )
        return mpe

    def _chisq(self, index, star_dict, err_dict, 
        err='average', return_star=False
    ):
        '''Convenience routine to compute the chi-squared of a fit.
        '''
        star = self.get_star_eep(index)

        chisq = 0
        for l in star_dict:
            if isinstance(err_dict[l], (tuple, list, np.array)):
                if err == 'average':
                    uncert = np.average(err_dict[l])
                elif err == 'min':
                    uncert = min(err_dict[l])
                elif err == 'max':
                    uncert = max(err_dict[l])
            else:
                uncert = err_dict[l]

            chisq += ((star[l] - star_dict[l]) / uncert)**2

        if return_star:
            return chisq, star
        return chisq

    def to_pickle(self, path=None):
        '''Saves the StarGridInterpolator to a pickle file.
        '''
        if path is None:
            path = os.path.join(grids_path, self.name, 'interpolator.pkl')
        with open(path, 'wb') as f:
            pickle.dump(self, f)

def altrange(start, stop, step):
    if stop == start:
        return [start]
    else:
        return np.arange(start, stop, step)

def _eep_pool_helper(tracks, eep_params, eep_functions, metric_function, i):
    return _eep_interpolate(tracks.loc[i], eep_params, eep_functions, metric_function)  

def load_interpolator(name=None, path=None):
    '''
    Load StarGridInterpolator from pickle file. If the interpolator has been
    cached during the install, simply specifying the name will be enough to
    load it.
    '''

    if name and path:
        raise ValueError('Please specify only `name` or `path`.')
    elif name:
        path = os.path.join(grids_path, name, 'interpolator.pkl')
        if not os.path.exists(path):
            path = os.path.join(interp_path, f'{name}.pkl')
            if not os.path.exists(path):
                raise FileNotFoundError(f"No interpolator found named '{name}'")
    elif not path:
        raise ValueError('Specify `name` or `path`.')
    
    with open(path, 'rb') as f:
        interp = pickle.load(f)
    return interp

def from_pandas(df, *args, **kwargs):
    '''Convert pandas DataFrame to StarGrid object.
    '''
    return StarGrid(df, *args, **kwargs)

def from_pickle(*args, **kwargs):
    '''Read StarGrid from pickle file.
    '''
    name = kwargs.pop('name', None)
    df = pd.read_pickle(*args, **kwargs)
    return from_pandas(df, name=name)

def from_csv(*args, **kwargs):
    '''Read StarGrid from csv.
    '''
    name = kwargs.pop('name', None)
    df = pd.read_csv(*args, **kwargs)
    return from_pandas(df, name=name)

def from_parquet(*args, **kwargs):
    '''
    Read StarGrid from parquet. Requires installation of pyarrow or
    similar package to support parquet.
    '''
    name = kwargs.pop('name', None)
    eep_params = kwargs.pop('eep_params', None)
    df = pd.read_parquet(*args, **kwargs)
    return from_pandas(df, name=name, eep_params=eep_params)

def install_grid(script, kind='raw'):
    '''
    Installs grid from a user-created setup file. For examples, see
    *_install.py scripts in the kiauhoku parent directory.

    Installation file must have a variable `name` set to the name of the grid,
    as well as a function `setup` that returns the set of stellar model grids
    as a MultiIndexed pandas DataFrame.

    Usage for `rotevol_install.py`:
        >>> import kiauhoku as kh
        >>> kh.install_grid('rotevol_install')
    '''

    # For now, MIST only works if the grids are already in EEP basis.
    if 'mist' in script and kind == 'raw':
        raise NotImplementedError(
            'For now, MIST input grids must already be in EEP basis.\n'
            'Please specify kind="eep".'
        )

    module = import_module(script)
    print(f'Installing grid "{module.name}" from {script}')

    # Create cache directories
    path = os.path.join(grids_path, module.name)
    if not os.path.exists(path):
        os.makedirs(path)

    if kind == 'raw':
        eep_params = module.eep_params
        # Cache eep parameters
        with open(os.path.join(path, 'eep_params.pkl'), 'wb') as f:
            pickle.dump(eep_params, f)

        print('Reading and combining grid files')
        grids = module.setup()
        grids = from_pandas(grids, name=module.name)

        # Save full grid to file
        full_save_path = os.path.join(path, 'full_grid.pqt')
        print(f'Saving to {full_save_path}')
        grids.to_parquet(full_save_path)

        print(f'Converting to eep-based tracks')
        try:
            eep_functions = module.eep_functions
        except AttributeError:
            eep_functions = None
        try:
            metric_function = module.metric_function
        except AttributeError:
            metric_function = None

        eeps = grids.to_eep(eep_params, eep_functions, metric_function)

    elif kind == 'eep':
        eeps = module.setup()
        eeps = from_pandas(eeps, name=module.name)

    # Save EEP grid to file
    eep_save_path = os.path.join(path, 'eep_grid.pqt')
    print(f'Saving to {eep_save_path}')
    eeps.to_parquet(eep_save_path)

    # Create and save interpolator to file
    interp = eeps.to_interpolator()
    interp_save_path = os.path.join(path, 'interpolator.pkl')
    print(f'Saving interpolator to {interp_save_path}')
    interp.to_pickle(path=interp_save_path)

    print(f'Model grid "{module.name}" installed.')

def load_full_grid(path=None, name=None):
    '''Load raw model grid from file.
    '''
    return load_grid(path=path, name=name, kind='full')

def load_eep_grid(path=None, name=None):
    '''Load EEP-based model grid from file.
    '''
    return load_grid(path=path, name=name, kind='eep')

def load_grid(path=None, name=None, kind='full'):
    '''Load model grid from file.
    '''
    if path:
        file_path = path
        if name:
            print('`kiauhoku.stargrid.load_grid`: `path` is specified; ignoring `name`.')
        else:
            name = os.path.basename(os.path.dirname(path))
    elif name:
        file_path = os.path.join(grids_path, name, f'{kind}_grid.pqt')

    if (kind == 'eep') or ('eep' in file_path):
        try:
            eep_params = load_eep_params(name)
        except:
            eep_params = None
    else:
        eep_params = None

    if os.path.exists(file_path):
        return from_parquet(file_path, name=name, eep_params=eep_params)
    raise FileNotFoundError(f"No such file or directory: '{file_path}'")

def load_eep_params(name):
    '''
    Assuming EEP params were specified in the setup script and cached,
    this will load them from the cache by specifying the grid name.
    '''
    params_path = os.path.join(grids_path, name, 'eep_params.pkl')
    with open(params_path, 'rb') as f:
        eep_params = pickle.load(f)

    return eep_params
